import os

import torch
import torch.nn as nn
import torch.nn.functional as F
from torchvision import transforms
from torch.utils.data import DataLoader
import gc

import utils
import utils.few_shot as fs
from config import Config
from imageClassificationAdapter import ImageClassifierAdapter
from datasets import WrapDataset, CategoriesSampler
from models import Classifier, MetaBaseline


class ImageClassifierAlgorithm(ImageClassifierAdapter):

    def __init__(self, toolset):
        # def __init__(self, problem, base_dataset, adapt_dataset, arguments):
        ImageClassifierAdapter.__init__(self, toolset)

    def _make_transform(self, name, image_size=None):
        # norm_params = {'mean': [0.485, 0.456, 0.406],
        #               'std': [0.229, 0.224, 0.225]}
        norm_params = {'mean': [0.5, 0.5, 0.5],
                       'std': [0.5, 0.5, 0.5]}  ##
        normalize = transforms.Normalize(**norm_params)  ##

        if name == 'identity':
            return transforms.Compose([
                transforms.ToTensor(),
                normalize,
            ])
        elif name == 'resize':
            return transforms.Compose([
                transforms.RandomResizedCrop(image_size),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
                normalize,
            ])
        else:
            raise NotImplementedError

    @staticmethod
    def _get_dataset_labels(dataset):
        ret = []
        for i in range(len(dataset)):
            ret.append(int(dataset[i][1]))
        return ret

    def initialize(self):
        """  Initialize your algorithm and save any parameters or objects you want
        to be persistent between checkpoints.

        """

        # pre-trained network on source task (the closest task)
        #  You are not required to use it but strongly suggested to use it
        #  These are in the style of torchvision models
        self.source_backbone = self.toolset['source_network']

        # Load the default configuration from config.py for your algorithm
        self.config = Config()

        # This sends the configuration back to the toolset
        self.toolset['classification_config'] = self.config

        self.image_size = self.toolset['source_dataset'][0][0].shape[1]

        # _gpus = os.environ['CUDA_VISIBLE_DEVICES']
        _gpus = self.config['gpu_ids']
        self._parallel = (len(_gpus) > 1)

        labeled_dataset = WrapDataset(
            self.toolset['source_dataset'],
            indices=self.toolset['source_dataset'].get_labeled_indices(),
            transform=self._make_transform('identity')
        )
        self.n_base_classes = max(self._get_dataset_labels(labeled_dataset)) + 1

        self.classification_pretraining()

        self.meta_learning()

    def classification_pretraining(self):
        #### Dataset ####

        labeled_dataset = WrapDataset(
            self.toolset['source_dataset'],
            indices=self.toolset['source_dataset'].get_labeled_indices(),
            transform=self._make_transform(
                self.config['cls_transform'], self.image_size)
        )
        labeled_dataloader = DataLoader(
            labeled_dataset,
            batch_size=self.config['cls_batch_size'],
            shuffle=True,
            num_workers=self.config['num_workers'],
            collate_fn=self.toolset['source_dataset'].collate_batch,
            drop_last=False,
        )

        ########

        #### Model and Optimizer ####
        # TODO: update when feat dim is set by incoming model
        feat_dim = 1000
        n_classes = self.n_base_classes
        model = Classifier(self.source_backbone, feat_dim, n_classes)

        if self.config['cuda']:
            model.cuda()
            if self._parallel:
                model = nn.DataParallel(model)

        optimizer, lr_scheduler = utils.make_optimizer(
            model.parameters(), 'sgd', **{
                'lr': self.config['cls_lr'],
                'weight_decay': 1.e-4,
                'milestones': self.config['cls_milestones']
            }
        )

        ########

        train_loader = labeled_dataloader
        max_epoch = self.config['cls_max_epoch']

        print('start classification pre-training ...')

        for epoch in range(1, max_epoch + 1):
            tl, ta = utils.Averager(), utils.Averager()

            # train
            model.train()
            for data, label in train_loader:
                if self.config['cuda']:
                    data, label = data.cuda(), label.cuda()
                logits = model(data)
                loss = F.cross_entropy(logits, label)
                acc = utils.compute_acc(logits, label)
                tl.add(loss.item())
                ta.add(acc)

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

                logits = None
                loss = None

            # post
            if lr_scheduler is not None:
                lr_scheduler.step()

            print('epoch {}, train: loss={:.4f} acc={:.4f}'.format(
                epoch, tl.item(), ta.item()))

        if self._parallel:
            model = model.module
        self.encoder = model.encoder.cpu()

    def meta_learning(self):

        #### Dataset ####

        labeled_dataset = WrapDataset(
            self.toolset['source_dataset'],
            indices=self.toolset['source_dataset'].get_labeled_indices(),
            transform=self._make_transform(
                self.config['meta_transform'], self.image_size)
        )
        label_list = self._get_dataset_labels(labeled_dataset)

        n_way, n_shot = 5, 5  ##
        n_query = 15  ##

        n_train_way = n_way
        n_train_shot = n_shot
        ep_per_batch = self.config['meta_batch_size']

        train_dataset = labeled_dataset
        train_sampler = CategoriesSampler(
            label_list, self.config['meta_epoch_size'],
            n_train_way, n_train_shot + n_query,
            ep_per_batch=ep_per_batch)
        train_loader = DataLoader(train_dataset, batch_sampler=train_sampler,
                                  num_workers=self.config['num_workers'])
        ########

        # ### Model and optimizer ####

        model = MetaBaseline(self.encoder)

        if self.config['cuda']:
            model.cuda()
            if self._parallel:
                model = nn.DataParallel(model)

        optimizer, lr_scheduler = utils.make_optimizer(
            model.parameters(), 'sgd', **{
                'lr': self.config['meta_lr'],
                'weight_decay': 1.e-4
            }
        )

        ########

        max_epoch = self.config['meta_max_epoch']

        print('start meta-learning ...')

        for epoch in range(1, max_epoch + 1):
            tl, ta = utils.Averager(), utils.Averager()

            # train
            model.train()
            if self.config['meta_freeze_bn']:
                utils.freeze_bn(model)

            for data, _ in train_loader:
                if self.config['cuda']:
                    data = data.cuda()
                x_shot, x_query = fs.split_shot_query(
                    data, n_train_way, n_train_shot, n_query,
                    ep_per_batch=ep_per_batch)
                label = fs.make_nk_label(n_train_way, n_query,
                                         ep_per_batch=ep_per_batch)
                if self.config['cuda']:
                    label = label.cuda()

                logits = model(x_shot, x_query).view(-1, n_train_way)
                loss = F.cross_entropy(logits, label)
                acc = utils.compute_acc(logits, label)
                tl.add(loss.item())
                ta.add(acc)

                optimizer.zero_grad()
                loss.backward()
                optimizer.step()

                logits = None
                loss = None

            print('epoch {}, meta-train: loss={:.4f} acc={:.4f}'.format(
                epoch, tl.item(), ta.item()))

        if self._parallel:
            model = model.module
        self.encoder = model.encoder.cpu()

    def compute_protos(self):
        labeled_dataset = WrapDataset(
            self.toolset['target_dataset'],
            indices=self.toolset['target_dataset'].get_labeled_indices(),
            transform=self._make_transform('identity')
        )
        labeled_dataloader = DataLoader(
            labeled_dataset,
            batch_size=self.config['eval_batch_size'],
            num_workers=self.config['num_workers'],
            collate_fn=self.toolset['target_dataset'].collate_batch,
            drop_last=False,
        )

        encoder = self.encoder
        encoder.eval()
        if self._parallel:
            encoder = nn.DataParallel(encoder)

        protos = dict()
        cnt = dict()
        n_classes = -1
        if self.config['cuda']:
            encoder = encoder.cuda()

        with torch.no_grad():

            for data, label in labeled_dataloader:
                if self.config['cuda']:
                    data = data.cuda()
                feat = encoder(data)
                for i, lb in enumerate(label):
                    lb = int(lb)
                    if protos.get(lb) is None:
                        protos[lb] = feat[i]
                        cnt[lb] = 1
                    else:
                        protos[lb] += feat[i]
                        cnt[lb] += 1
                    if lb + 1 > n_classes:
                        n_classes = lb + 1

            _protos = []
            for c in range(n_classes):
                _protos.append(protos[c] / cnt[c])
            self.protos = torch.stack(_protos)
            self.protos = F.normalize(self.protos, dim=-1)

    def domain_adapt_training(self):
        """ This is where you will fine-tune/train your few-shot or domain adaption
        algorithm.  You may use the dataloaders below and feel free to override the
        transformers in the datasets to use whatever pretraining you desire.
        Please do not modify the datasets though.
        """
        self.compute_protos()

        gc.collect()  # Make sure all object have ben deallocated if not used
        torch.cuda.empty_cache()

    def inference(self):
        """
        Inference is during the evaluation stage.  For this example, the
        task_network is trained in the train and adapt stage's code and
        is used here to create the predictions.  The indices are used to
        track the images/filenames for submitting the predictions back to
        JPL

        Returns:
            tuple(list[int],list[int]): preds and indices
                predicted category indices and image indices

        """
        # self.task_model.eval()
        # eval_dataloader = data.DataLoader(
        #     toolset['Dataset'],
        #     batch_size=int(self.arguments['batch_size']),
        #     drop_last=False
        # )
        # preds = []
        # indices = []
        #
        # for imgs, inds in eval_dataloader:
        #     if self.cuda:
        #         imgs = imgs.cuda()
        #
        #     with torch.no_grad():
        #         preds_ = self.task_model(imgs)
        #
        #     preds += torch.argmax(preds_, dim=1).cpu().numpy().tolist()
        #     indices += inds.numpy().tolist()

        encoder = self.encoder

        if self.config['cuda']:
            encoder = encoder.cuda()
        encoder.eval()
        if self._parallel:
            encoder = nn.DataParallel(encoder)

        eval_dataset = WrapDataset(
            self.toolset['eval_dataset'],
            transform=self._make_transform('identity'),
            is_eval=True
        )
        eval_dataloader = DataLoader(
            eval_dataset,
            batch_size=self.config['eval_batch_size'],
            drop_last=False
        )

        preds = []
        indices = []
        with torch.no_grad():
            for data, inds in eval_dataloader:
                if self.config['cuda']:
                    data = data.cuda()
                feat = self.encoder(data)
                feat = F.normalize(feat, dim=-1)
                logits = torch.mm(feat, self.protos.t())
                preds += torch.argmax(logits, dim=-1).cpu().tolist()
                indices += inds.tolist()


        gc.collect()  # Make sure all object have ben deallocated if not used
        torch.cuda.empty_cache()

        return preds, indices
