from framework.localinterface import LocalInterface
from framework.baseprotocol import BaseProtocol

# Note: doing full include to allow for documentation.  This isn't necessary
#     for just running.
from learn.protocol.learn_config import LearnConfig


class Learn(BaseProtocol):
    """
        The protocol class should derive from two source classes. The first is the
        BaseProtocol. This provides the self.get_algorithm function for locating and
        creating an algorithm object for you given only the filename of the file
        containing the algorithm. To use the functions of that algorithm, call the
        execute() function. This call performs two critical actions that are not
        available by directly calling the algorithms functions. First, it establishes
        an indirect call path that allows the automated tools to generate and
        annotate the adapter and template files for use when creating brand new
        algorithms.  Second, it allows the deconstruction of the toolset dict(), and
        automatically stores the toolset for local use within the algorithm itself.

        The second class that the protocol class should derive from is the desired
        back-end data and reporting structure. One such back-end is the
        JPLInterface which provides access to the JPL T&E server for the learn
        project (and hopefully sail-on as well). Another useful back-end is the
        LocalInterface which provides similar access as the JPLInterface using
        local data only. The interface mimics the JPLInterface, but works entirely
        within the framework, and requires no external server to function.

        Only Kitware should be modifying this function.

    """
    def __init__(self, algorithms_directory, harness):
        BaseProtocol.__init__(self, algorithms_directory, harness)

        self.config = LearnConfig()
        self.toolset['protocol_config'] = self.config

    def run_protocol(self):
        """ run_protocol is called from the framework and represents that initial
            execution point of the protocol.
        """
        task_ids = self.test_harness.get_task_ids()
        tasks_to_run = self.config['tasks_to_run']
        print('List of Potential Tasks:')
        print(f'{task_ids}\n\n')
        if len(tasks_to_run) < 1:
            print('Running All Tasks')
            tasks_to_run = task_ids
        else:
            print('Running following tasks:')
        print(tasks_to_run)

        for task in tasks_to_run:
            if self.test_harness.get_problem_metadata(task)['problem_type'] \
                    == 'machine_translation':
                continue
            try:
                self.run_task(task)
            finally:
                self.test_harness.terminate_session()

    def run_task(self, task_id: str) -> None:
        """

        Args:
            task_id (str): task id for which you want to run.

        Returns:

        """
        self.test_harness.initialize_session(task_id)
        self.toolset["problem_type"] = self.test_harness.metadata['problem_type']
        self.toolset["whitelist_datasets"] = self.test_harness.get_whitelist_datasets_jpl()

        domain_select_algo = self.get_algorithm(
            self.config['domain_network_selection_algo'], self.toolset)

        algo_select_algo = self.get_algorithm(self.config['algoSelection_algo'],
                                              self.toolset)

        stages = self.test_harness.get_stages()

        for stage in stages:
            # self.toolset["target_dataset"] = self.get_dataset(stage, 'train')
            self.toolset["target_dataset"] = self.test_harness.get_dataset_jpl(
                stage, 'train')
            self.toolset["eval_dataset"] = self.test_harness.get_dataset_jpl(
                stage, 'eval', categories=self.toolset['target_dataset'].categories)

            source_network, source_dataset = domain_select_algo.execute(
                self.toolset, "SelectNetworkAndDataset")

            self.toolset["source_dataset"] = source_dataset
            self.toolset["source_network"] = source_network

            query_algo_id, adapt_algo_id = algo_select_algo.execute(
                self.toolset, "SelectAlgorithms")

            query_algo = self.get_algorithm(query_algo_id, self.toolset)
            query_algo.execute(self.toolset, "Initialize")

            # check if we are supposed to use the same source for both
            # query and adapt. if not, then init the adapt separately.
            if adapt_algo_id == query_algo_id:
                adapt_algo = query_algo
            else:
                adapt_algo = self.get_algorithm(adapt_algo_id, self.toolset)
                adapt_algo.execute(self.toolset, "Initialize")

            checkpoints = self.test_harness.get_budget_checkpoints(
                stage, self.toolset["target_dataset"]
            )

            for it_cp, cp in enumerate(checkpoints):
                self.test_harness.start_checkpoint(
                    stage, self.toolset["target_dataset"], it_cp)
                self.toolset["budget"] = self.test_harness.get_remaining_budget()
                # call the query_algo to update the dataset with new labels
                if self.toolset["budget"] > 0:
                    query_algo.execute(self.toolset, "SelectAndLabelData")

                # call the adapt_algo to update the model to incorporate the
                # new labels
                adapt_algo.execute(self.toolset, "DomainAdaptTraining")
                results = adapt_algo.execute(self.toolset, "EvaluateOnTestDataSet")
                self.test_harness.post_results(
                    stage, self.toolset["eval_dataset"], results)

            # Add stage dataset to the whitelist
            train_name = f'{stage}_{self.toolset["target_dataset"].name}_train'
            self.toolset["whitelist_datasets"][train_name] = self.toolset["target_dataset"]
            test_name = f'{stage}_{self.toolset["eval_dataset"].name}_test'
            self.toolset["whitelist_datasets"][test_name] = self.toolset["eval_dataset"]
            print(self.test_harness.format_status(False))
        self.test_harness.terminate_session()
